//############################################################################
//
// EZFB ~ Linux Frame Buffer API ~
//
// by James Lehman
// james@akrobiz.com
//
// began: Feb. 2001
//
// EZFB is intended to give C programmers an easy-to-use library of functions
// for drawing points, lines and strings of characters into the Linux frame
// buffer. It also provides a means of displaying bitmap images on any portion
// of the screen and capturing the contents of any portion of the screen to
// a standard bitmap graphics file. All of this is embodied in a set of
// functions that are polymorphic with regard to the characteristics of the
// frame buffer and the bitmap files used. In other words, it makes no
// difference what the screen height or width is or what the color depth is,
// all of the function calls are the same and will have appropriate and
// predictable effects.
//
// This software is distributed in the hope that it will be useful, but
// without any warranty; without even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public License
// for more details.
//
//############################################################################

#include "ezfb_tty.h"
#include "ezfb_colors.h"
#include "ezfb_device.h"

//############################################################################
int    open_fb         (struct ezfb* fb, char* dev_name);
int    get_fb_var      (struct ezfb* fb);
int    set_fb_var      (struct ezfb* fb);
int    get_fb_fix      (struct ezfb* fb);
int    get_fb_mem      (struct ezfb* fb, int save_old);
int    find_screen     (struct ezfb* fb);
int    init_ezfb_base  (struct ezfb* fb, char* dev_name, int save_old);
int    copy_old_fb     (struct ezfb* fb);
void   set_sig_traps   ();
void   terminate       (int signum);

//############################################################################
struct ezfb* valid_ezfb[MAX_FB_NUM + 1] = {NULL};
u_int        num_valid_ezfb             = 0;

//############################################################################
int fb_dev_exists(char* dev_name)
{
	int fd;
EZFB_FUNCTION_CALL_1
    if(0 > (fd = open(dev_name, O_RDWR)))
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: FRAME BUFFER DOES NOT EXIST");
        printf("(%s)", dev_name);
#endif
        ret = 0;
    }
    else
    {
#if EZFB_API_MESSAGES
        printf("\nFRAME BUFFER %s EXSISTS", dev_name);fflush(stdout);
		close(fd);
#endif
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int fb_num_exists(int num)
{
	char dev_name[16] = {0};
EZFB_FUNCTION_CALL_1
	strcpy(dev_name, FB_DEV_NAME);
	sprintf(&(dev_name[strlen(dev_name)]), "%d", num);
	ret = fb_dev_exists(dev_name);
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
u_int fb_name_to_num(char* dev_name) // returns last one or two digits as u_int
{
    if(isdigit(dev_name[strlen(dev_name) - 2]))
        return (u_int)atoi(&dev_name[strlen(dev_name) - 2]);
    return (u_int)atoi(&dev_name[strlen(dev_name) - 1]);
}

//############################################################################
int open_fb(struct ezfb* fb, char* dev_name)
{
EZFB_FUNCTION_CALL_1
    if(0 > (fb->fd = open(dev_name, O_RDWR)))
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: OPENING FRAME BUFFER DEVICE FAILED");
        printf("(%s)", dev_name);
#endif
        ret = 0;
    }
    else
    {
        strcpy(fb->dev_name, dev_name);
        fb->dev_num = fb_name_to_num(fb->dev_name);
#if EZFB_API_MESSAGES
        printf("\nFRAME BUFFER %s SUCCESSFULLY OPENED", fb->dev_name);fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int get_fb_var(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_0
    if(0 > ioctl(fb->fd, FBIOGET_VSCREENINFO, &fb->Var))
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: FBIOGET_VSCREENINFO failed");
#endif
    }
    else // ioctl did not fail
    {
        if(!fb->exists)
        {
            fb->bpp_was          = fb->Var.bits_per_pixel;
            //fb->xres_was         = fb->Var.xres;
            //fb->yres_was         = fb->Var.yres;
            //fb->xres_virtual_was = fb->Var.xres_virtual;
            //fb->yres_virtual_was = fb->Var.yres_virtual;
        }
        fb->max_x = fb->Var.xres - 1;
        fb->max_y = fb->Var.yres - 1;
        ret       = 1;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int set_fb_var(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(0 > ioctl(fb->fd, FBIOPUT_VSCREENINFO, &fb->Var))
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: FBIOPUT_VSCREENINFO failed");
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int get_fb_fix(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(0 > ioctl(fb->fd, FBIOGET_FSCREENINFO, &fb->Fix))
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: FBIOGET_FSCREENINFO failed");
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int get_fb_mem(struct ezfb* fb, int save_old)
{
EZFB_FUNCTION_CALL_1
    if(0 > (int)(fb->data = (u_char*)mmap(NULL, fb->Fix.smem_len, (PROT_READ | PROT_WRITE), MAP_SHARED, fb->fd, 0)))
    {
#if EZFB_API_ERRORS
        perror("\nezfb ERROR: MEMORY MAP FOR FRAME BUFFER DEVICE FAILED");
#endif
        ret = 0;
    }
    else if(save_old)
    {
        fb->data_was = (u_char*)malloc(fb->Fix.smem_len * sizeof(u_char));
        printf("\nSAVING CONTENTS OF VIDEO RAM...\na lot of video memory takes a while.");fflush(stdout);
        memcpy((void*)fb->data_was, (const void*)fb->data, fb->Fix.smem_len); // this can take a while!
#if EZFB_API_MESSAGES
        printf("\nCONTENTS OF VIDEO RAM SAVED");fflush(stdout);
#endif
    }
    else
    {
        fb->data_was = NULL; // just to be sure
#if EZFB_API_MESSAGES
        printf("\nCONTENTS OF VIDEO RAM NOT SAVED");fflush(stdout);
#endif
    }

EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int find_screen(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
        fb->screen_uints    = (u_int*)
        fb->screen_ushorts  = (u_short*)
        fb->screen_uchars   = fb->data + (fb->Var.yoffset * fb->Var.xres_virtual);
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int init_ezfb_base(struct ezfb* fb, char* dev_name, int save_old)
{
EZFB_FUNCTION_CALL_1
    srandom(time(NULL));
    set_sig_traps();
    if(fb)
    {
        if(valid_ezfb[fb_name_to_num(dev_name)] == NULL)
        {
            if(!fb->exists)
            {
                ret = open_fb(fb, dev_name); // sets fb->dev_name, fb->dev_num
                if(ret)                      ret &= get_fb_fix(fb);
                if(ret)                      ret &= get_fb_var(fb);
                if(ret)                      ret &= get_fb_mem(fb, save_old);
                if(ret)                      ret &= init_ezfb_cmap(fb, save_old);
                if(ret && fb->dev_num == 0)  ret &= set_tty_graphics();
                if(ret)                      fb->exists = 1;
                if(ret)                      ret &= find_screen(fb);
                if(ret)                      valid_ezfb[fb->dev_num] = fb;
	            if(ret)                      num_valid_ezfb++;
                if(ret)
                {
                    fb->colors.black  = 0x00000000;
                    fb->colors.white  = 0x00ffffff;
                }
#if EZFB_API_MESSAGES
                if(ret)
                    printf("\nezfb %s SUCCESSFULLY INITIALIZED", fb->dev_name);fflush(stdout);
#endif
            } // if(!fb->exists)
            else
            {
    #if EZFB_API_ERRORS
                fprintf(stderr, "\nezfb ERROR: init_ezfb( %s ) CALLED MORE THAN ONCE!", fb->dev_name);fflush(stderr);
    #endif
                ret = 0;
            }
        } // if(valid_ezfb[fb_name_to_num(dev_name)] == NULL)
        else
        {
#if EZFB_API_ERRORS
            fprintf(stderr, "\nezfb ERROR: FRAME BUFFER %s ALREADY OPEN BY THIS APPLICATION", dev_name);fflush(stderr);
#endif
            ret = 0;
        }
    } //if(fb)
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: CAN NOT INITIALIZE NULL POINTER");fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int init_ezfb(struct ezfb* fb) // each call attmepts to init next fb
{
EZFB_FUNCTION_CALL_1
    if(num_valid_ezfb <= MAX_FB_NUM)
        ret = init_ezfb_num(fb, num_valid_ezfb);
#if EZFB_API_ERRORS
    if(!ret)
        fprintf(stderr, "\nezfb ERROR: NO MORE VALID FRAME BUFFER DEVICES");fflush(stderr);
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int init_ezfb_no_save(struct ezfb* fb) // each call attmepts to init next fb
{
EZFB_FUNCTION_CALL_1
    if(num_valid_ezfb <= MAX_FB_NUM)
        ret = init_ezfb_num_no_save(fb, num_valid_ezfb);
#if EZFB_API_ERRORS
    if(!ret)
        fprintf(stderr, "\nezfb ERROR: NO MORE VALID FRAME BUFFER DEVICES");fflush(stderr);
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int init_ezfb_num(struct ezfb* fb, u_int dev_num)
{
    char dev_name[MAX_LENGTH_OF_DEV_NAME] = {'\0'};
EZFB_FUNCTION_CALL_1
    if(dev_num >= 0 && dev_num <= MAX_FB_NUM)
    {
        sprintf(dev_name, "%s%u", FB_DEV_NAME, dev_num);
        if(!(ret = init_ezfb_base(fb, dev_name, 1)))
        {
            sprintf(dev_name, "%s/%d", FB_DEV_NAME, dev_num);
            ret = init_ezfb_base(fb, dev_name, 1);
        }
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: %d OUT OF RANGE!", dev_num);fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int init_ezfb_num_no_save(struct ezfb* fb, u_int dev_num)
{
    char dev_name[MAX_LENGTH_OF_DEV_NAME] = {'\0'};
EZFB_FUNCTION_CALL_1
    if(dev_num >= 0 && dev_num <= MAX_FB_NUM)
    {
        sprintf(dev_name, "%s%u", FB_DEV_NAME, dev_num);
        if(!(ret = init_ezfb_base(fb, dev_name, 0)))
        {
            sprintf(dev_name, "%s/%d", FB_DEV_NAME, dev_num);
            ret = init_ezfb_base(fb, dev_name, 0);
        }
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: %d OUT OF RANGE!", dev_num);fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int init_named_ezfb(struct ezfb* fb, char* dev_name)
{
EZFB_FUNCTION_CALL_1
    if(!init_ezfb_base(fb, dev_name, 1))
        ret = 0;
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int init_named_ezfb_no_save(struct ezfb* fb, char* dev_name)
{
EZFB_FUNCTION_CALL_1
    if(!init_ezfb_base(fb, dev_name, 0))
        ret = 0;
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int set_ezfb_bpp(struct ezfb* fb, u_int bpp_wanted)
{
EZFB_FUNCTION_CALL_0
    if(fb->exists)
    {
        if(fb->Var.bits_per_pixel == bpp_wanted)
        {
#if EZFB_API_FUNCTION_NIT_PICKING
            fprintf(stderr, "\nezfb ERROR: ALREADY AT %d bpp.", bpp_wanted);fflush(stderr);
#endif
            ret = 1; // error, but return ok
        }
        else
        {
            fb->Var.bits_per_pixel = bpp_wanted;
            ret  = set_fb_var (fb);
            ret &= get_fb_var (fb);
            if(fb->Var.bits_per_pixel != bpp_wanted) // bpp not supported
            {
#if EZFB_API_ERRORS
                fprintf(stderr, "\nezfb ERROR: set_ezfb_bpp(fb, %d) FAILED", bpp_wanted);fflush(stderr);
#endif
                ret = 0;
            }
            else
            {
                fb->colors.fbcmap.len = (fb->Var.bits_per_pixel <= 8)
                                           ? (0x01 << fb->Var.bits_per_pixel)
                                           : 0;
                ret &= set_ezfb_cmap        (fb);
                ret &= get_ezfb_cmap        (fb);
                ret &= get_color_factors    (fb);
                if(fb->Var.bits_per_pixel <= 8)
                    ret &= set_ezfb_cmap_utility (fb);
                ret &= find_black_and_white (fb);
                ret &= find_screen          (fb);
                if(ret && (fb->Var.bits_per_pixel != fb->bpp_was))
                    fb->bpp_was_changed = 1;
                else if(fb->Var.bits_per_pixel == fb->bpp_was)
                    fb->bpp_was_changed = 0;
            }
        }
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (set_ezfb_bpp) FB NOT INITIALIZED");fflush(stderr);
#endif
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
//int set_ezfb_resolution(struct ezfb* fb, int new_xres, int new_yres)
//{       //TODO: make this work!
//EZFB_FUNCTION_CALL_1
//    if(fb->exists)
//    {
//        fb->Var.xres = new_xres;
//        fb->Var.yres = new_yres;
//        ret &= set_fb_var(fb);
//        ret &= get_fb_var(fb);
//        if(ret) fb->resolution_changed = 1;
//    }
//    else
//    {
//#if EZFB_API_ERRORS
//        fprintf(stderr, "\nezfb ERROR: (set_ezfb_resolution) FB NOT INITIALIZED");fflush(stderr);
//#endif
//        ret = 0;
//    }
//EZFB_FUNCTION_RETURN(ret)
//}

//############################################################################
int copy_old_fb(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(fb->exists && fb->data_was)
    {   // TODO: deal with x & y offsets that might have changed
        memcpy((void*)fb->data, (const void*)fb->data_was, fb->Fix.smem_len);
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: fb->data_was NEVER SAVED!");fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int clear_ezfb(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(fb->exists)
    {
        memset((void*)fb->data, 0x00, fb->Fix.smem_len);
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (clear_ezfb) FB NOT INITIALIZED");fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int clear_screen(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_1
    if(fb->exists)
    {
        find_screen(fb);
        memset((void*)fb->screen_uchars, 0, (fb->Var.xres * fb->Var.yres * 8) / fb->Var.bits_per_pixel);
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (clear_screen) FB NOT INITIALIZED");fflush(stderr);
#endif
        ret = 0;
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int release_ezfb(struct ezfb* fb)
{
EZFB_FUNCTION_CALL_0
    if(fb)
    {
        if(fb->exists)
        {
            ret = 1;
            if(fb->bpp_was_changed)
                ret &= set_ezfb_bpp(fb, fb->bpp_was);
            if(fb->data_was)
            {
                memcpy((void*)fb->data, (const void*)fb->data_was, fb->Fix.smem_len); // TODO reset the x & y offsets
                free(fb->data_was);
                fb->data_was = NULL;
            }
            ret &= release_ezfb_cmap(fb);
            if(fb->dev_num == 0)
                ret &= set_tty_text();
            close(fb->fd);
            fb->exists = 0;
            valid_ezfb[fb->dev_num] = NULL;
		    num_valid_ezfb--;
#if EZFB_API_MESSAGES
            if(ret)
                printf("\nezfb %s SUCCESSFULLY RELEASED\n", fb->dev_name);fflush(stdout);
#endif
        }
        else
        {
#if EZFB_API_ERRORS
            fprintf(stderr, "\nezfb ERROR: (release_ezfb) FB NOT INITIALIZED");fflush(stderr);
#endif
        }
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: RELEASE CALLED ON NULL POINTER");fflush(stderr);
#endif
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int dump_fb_fix_screeninfo(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
#if EZFB_API_MESSAGES
        printf("\n      FB FIXED SCREEN INFO:"                                               );
        printf("\n      fb->Fix.id               = %s"  ,        fb->Fix.id           ); // identification string eg "TT Builtin"
        printf("\n      fb->Fix.smem_start       = 0x%X", (u_int)fb->Fix.smem_start   ); // Start of frame buffer mem (physical address)
        printf("\n      fb->Fix.smem_len         = %u"  ,        fb->Fix.smem_len     ); // Length of frame buffer mem
        printf("\n      fb->Fix.type             = %u"  ,        fb->Fix.type         ); // see FB_TYPE_*        */
        printf("\n      fb->Fix.type_aux         = %u"  ,        fb->Fix.type_aux     ); // Interleave for interleaved Planes
        printf("\n      fb->Fix.visual           = %u"  ,        fb->Fix.visual       ); // see FB_VISUAL_*        */
        printf("\n      fb->Fix.xpanstep         = %u"  ,        fb->Fix.xpanstep     ); // zero if no hardware panning
        printf("\n      fb->Fix.ypanstep         = %u"  ,        fb->Fix.ypanstep     ); // zero if no hardware panning
        printf("\n      fb->Fix.ywrapstep        = %u"  ,        fb->Fix.ywrapstep    ); // zero if no hardware ywrap
        printf("\n      fb->Fix.line_length      = %u"  ,        fb->Fix.line_length  ); // length of a line in bytes
        printf("\n      fb->Fix.mmio_start       = 0x%X", (u_int)fb->Fix.mmio_start   ); // Start of Memory Mapped I/O (physical address)
        printf("\n      fb->Fix.mmio_len         = %u"  ,        fb->Fix.mmio_len     ); // Length of Memory Mapped I/O
        printf("\n      fb->Fix.accel            = %u"  ,        fb->Fix.accel        ); // Type of acceleration available
        printf("\n      fb->Fix.reserved[0]      = %u"  ,        fb->Fix.reserved[0]  ); // Reserved for future compatibility
        printf("\n      fb->Fix.reserved[1]      = %u"  ,        fb->Fix.reserved[1]  ); // Reserved for future compatibility
        printf("\n      fb->Fix.reserved[2]      = %u"  ,        fb->Fix.reserved[2]  ); // Reserved for future compatibility
        printf("\n\n");fflush(stdout);
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (dump_fb_fix_screeninfo) FB NOT INITIALIZED");fflush(stderr);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int dump_fb_var_screeninfo(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
#if EZFB_API_MESSAGES
        printf("\n      FB VARIABLE SCREEN INFO:"                                  );
        printf("\n      fb->Var.xres             = %u"  , fb->Var.xres             ); // visible resolution
        printf("\n      fb->Var.yres             = %u"  , fb->Var.yres             );
        printf("\n      fb->Var.xres_virtual     = %u"  , fb->Var.xres_virtual     ); // virtual resolution
        printf("\n      fb->Var.yres_virtual     = %u"  , fb->Var.yres_virtual     );
        printf("\n      fb->Var.xoffset          = %u"  , fb->Var.xoffset          ); // offset from virtual to visible
        printf("\n      fb->Var.yoffset          = %u"  , fb->Var.yoffset          ); // resolution
        printf("\n      fb->Var.bits_per_pixel   = %u"  , fb->Var.bits_per_pixel   ); // guess what
        printf("\n      fb->Var.grayscale        = %u"  , fb->Var.grayscale        ); // != 0 Graylevels instead of colors
        printf("\n      fb->Var.red.offset       = %u"  , fb->Var.red.offset       ); // bitfield in fb mem if true color,
        printf("\n      fb->Var.red.length       = %u"  , fb->Var.red.length       ); // else only length is significant
        printf("\n      fb->Var.red.msb_right    = %u"  , fb->Var.red.msb_right    );
        printf("\n      fb->Var.green.offset     = %u"  , fb->Var.green.offset     );
        printf("\n      fb->Var.green.length     = %u"  , fb->Var.green.length     );
        printf("\n      fb->Var.green.msb_right  = %u"  , fb->Var.green.msb_right  );
        printf("\n      fb->Var.blue.offset      = %u"  , fb->Var.blue.offset      );
        printf("\n      fb->Var.blue.length      = %u"  , fb->Var.blue.length      );
        printf("\n      fb->Var.blue.msb_right   = %u"  , fb->Var.blue.msb_right   );
        printf("\n      fb->Var.transp.offset    = %u"  , fb->Var.transp.offset    ); // transparency
        printf("\n      fb->Var.transp.length    = %u"  , fb->Var.transp.length    );
        printf("\n      fb->Var.transp.msb_right = %u"  , fb->Var.transp.msb_right );
        printf("\n      fb->Var.nonstd           = %u"  , fb->Var.nonstd           ); // != 0 Non standard pixel format
        printf("\n      fb->Var.activate         = %u"  , fb->Var.activate         ); // see FB_ACTIVATE_*
        printf("\n      fb->Var.height           = %u"  , fb->Var.height           ); // height of picture in mm
        printf("\n      fb->Var.width            = %u"  , fb->Var.width            ); // width of picture in mm
        printf("\n      fb->Var.accel_flags      = %u"  , fb->Var.accel_flags      ); // acceleration flags (hints)
        printf("\n      fb->Var.pixclock         = %u"  , fb->Var.pixclock         ); // pixel clock in ps (pico seconds)
        printf("\n      fb->Var.left_margin      = %u"  , fb->Var.left_margin      ); // time from sync to picture
        printf("\n      fb->Var.right_margin     = %u"  , fb->Var.right_margin     ); // time from picture to sync
        printf("\n      fb->Var.upper_margin     = %u"  , fb->Var.upper_margin     ); // time from sync to picture
        printf("\n      fb->Var.lower_margin     = %u"  , fb->Var.lower_margin     );
        printf("\n      fb->Var.hsync_len        = %u"  , fb->Var.hsync_len        ); // length of horizontal sync
        printf("\n      fb->Var.vsync_len        = %u"  , fb->Var.vsync_len        ); // length of vertical sync
        printf("\n      fb->Var.sync             = %u"  , fb->Var.sync             ); // see FB_SYNC_*
        printf("\n      fb->Var.vmode            = %u"  , fb->Var.vmode            ); // see FB_VMODE_*
        printf("\n      fb->Var.reserved[0]      = %u"  , fb->Var.reserved[0]      ); // Reserved for future compatibility
        printf("\n      fb->Var.reserved[1]      = %u"  , fb->Var.reserved[1]      ); // Reserved for future compatibility
        printf("\n      fb->Var.reserved[2]      = %u"  , fb->Var.reserved[2]      ); // Reserved for future compatibility
        printf("\n      fb->Var.reserved[3]      = %u"  , fb->Var.reserved[3]      ); // Reserved for future compatibility
        printf("\n      fb->Var.reserved[4]      = %u"  , fb->Var.reserved[4]      ); // Reserved for future compatibility
        printf("\n      fb->Var.reserved[5]      = %u"  , fb->Var.reserved[5]      ); // Reserved for future compatibility
        printf("\n\n");fflush(stdout);
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (dump_fb_var_screeninfo) FB NOT INITIALIZED");fflush(stderr);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int dump_ezfb(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
#if EZFB_API_MESSAGES
        printf("\n      EZFB GLOBAL PARAMETERS:"                                  );
        printf("\n      fb->fd                 = %d"  , fb->fd                    );
        printf("\n      fb->dev_num            = %u"  , fb->dev_num               );
        printf("\n      fb->dev_name           = %s"  , fb->dev_name              );
        printf("\n      fb->exists             = %u"  , fb->exists                );
        //printf("\n      fb->resolution_changed = %u"  , fb->resolution_changed    );
        printf("\n      fb->bpp_was_changed    = %u"  , fb->bpp_was_changed       );
        //printf("\n      fb->xres_was           = %u"  , fb->xres_was              );
        //printf("\n      fb->yres_was           = %u"  , fb->yres_was              );
        //printf("\n      fb->xres_virtual_was   = %u"  , fb->xres_virtual_was      );
        //printf("\n      fb->yres_virtual_was   = %u"  , fb->yres_virtual_was      );
        printf("\n      fb->bpp_was            = %u"  , fb->bpp_was               );
        printf("\n      fb->data               = 0x%X", (u_int)fb->data           );
        printf("\n      fb->screen_uchars      = 0x%X", (u_int)fb->screen_uchars  );
        printf("\n      fb->screen_ushorts     = 0x%X", (u_int)fb->screen_ushorts );
        printf("\n      fb->screen_uints       = 0x%X", (u_int)fb->screen_uints   );
        printf("\n      fb->data_was           = 0x%X", (u_int)fb->data_was       );
        printf("\n      fb->max_x              = %u"  , fb->max_x                 );
        printf("\n      fb->max_y              = %u"  , fb->max_y                 );
        printf("\n\n");fflush(stdout);
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (dump_ezfb) FB NOT INITIALIZED");fflush(stderr);
#endif
    }
EZFB_FUNCTION_RETURN(1)
}

//############################################################################
int dump_ezfb_screen_map(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
#if EZFB_API_MESSAGES
        u_int i;
        printf("\nFB SCREEN MAP %d x %d : %d bytes:", fb->Var.xres, fb->Var.yres, fb->Var.xres * fb->Var.yres);
        for(i = 0; i < fb->Var.xres * fb->Var.yres; i++)
        {
            if(0 == (i % fb->Var.xres))
                printf("\n");
            printf(" %2X" , fb->screen_uchars[i]);fflush(stdout);
        }
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (dump_ezfb_screen_map) FB NOT INITIALIZED");fflush(stderr);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int dump_ezfb_virtual_map(struct ezfb* fb)
{
EZFB_FUNCTION_CALL
    if(fb->exists)
    {
#if EZFB_API_MESSAGES
        u_int i;
        printf("\nFB VIRTUAL PIXEL MAP %d x %d : %d bytes:", fb->Var.xres_virtual, fb->Var.yres_virtual, fb->Var.xres_virtual * fb->Var.yres_virtual);
        for(i = 0; i < fb->Fix.smem_len; i++)
        {
            if(0 == (i % fb->Var.xres_virtual))
                printf("\n");
            printf(" %2X" , fb->data[i]);fflush(stdout);
        }
#endif
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (dump_ezfb_virtual_map) FB NOT INITIALIZED");fflush(stderr);
#endif
    }
EZFB_FUNCTION_RETURN(fb->exists)
}

//############################################################################
int binfile_ezfb_screen_map(struct ezfb* fb, char* file_name)
{
EZFB_FUNCTION_CALL_0
    if(fb->exists)
    {
        int binfile_fd;
        if(0 > (binfile_fd = open(file_name, (O_WRONLY | O_CREAT), (S_IRWXU | S_IRGRP | S_IROTH)))) // mode 744
        {
#if EZFB_API_ERRORS
            perror("\nezfb ERROR: opening output file failed");
#endif
            ret = 0;
        }
        else
        {
            ret = 1;
            write(binfile_fd, (void*)fb->screen_uchars, fb->Var.xres * fb->Var.yres);
            close(binfile_fd);
        }
    }
    else
    {
#if EZFB_API_ERRORS
        fprintf(stderr, "\nezfb ERROR: (binfile_ezfb_screen_map) FB NOT INITIALIZED");fflush(stderr);
#endif
    }
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
int binfile_ezfb_virtual_map(struct ezfb* fb, char* file_name)
{
EZFB_FUNCTION_CALL_0
    if(fb->exists)
    {
        int binfile_fd;
        if(0 > (binfile_fd = open(file_name, (O_WRONLY | O_CREAT), (S_IRWXU | S_IRGRP | S_IROTH)))) // mode 744
        {
#if EZFB_API_ERRORS
            perror("\nezfb ERROR: opening output file failed");
#endif
            ret = 0;
        }
        else
        {
            ret = 1;
            write(binfile_fd, (void*)fb->data, fb->Fix.smem_len);
            close(binfile_fd);
        }
    }
#if EZFB_API_ERRORS
    else
        fprintf(stderr, "\nezfb ERROR: (binfile_ezfb_virtual_map) FB NOT INITIALIZED");fflush(stderr);
#endif
EZFB_FUNCTION_RETURN(ret)
}

//############################################################################
float random_01() // returns a float between 0 & 1.
{
    //return ((float)(rand() % RAND_MAX) / (float)RAND_MAX);
    return ((float)(rand()) / (float)RAND_MAX);
}

//############################################################################
float random_neg_to_pos_1() // returns a float between -1 & 1.
{
    //return ((float)(rand() % RAND_MAX) / (float)RAND_MAX);
    return sin(rand());
}

//############################################################################
void set_sig_traps()
{
EZFB_FUNCTION_CALL
    static int been_done = 0;
    if(!been_done)
    {
        signal(SIGTERM, terminate);
        signal(SIGINT,  terminate);
        signal(SIGQUIT, terminate);
        signal(SIGHUP,  terminate);
        signal(SIGTSTP, terminate);
        signal(SIGTTIN, terminate);
        signal(SIGFPE,  terminate);
        signal(SIGILL,  terminate);
        signal(SIGSEGV, terminate);
        signal(SIGBUS,  terminate);
        signal(SIGABRT, terminate);
        signal(SIGSYS,  terminate);
        been_done = 1;
#if EZFB_API_MESSAGES
        printf("\nTERMINATION SIGNAL TRAPS SET");fflush(stdout);
#endif
    }
EZFB_FUNCTION_RETURN_
}

//############################################################################
void terminate(int signum)
{

    int i;
    for(i = 0; i <= MAX_FB_NUM; i++)
    {
        if(valid_ezfb[i])
        {
#if EZFB_API_MESSAGES
            printf("\nFRAME BUFFER %s TERMINATING... ", valid_ezfb[i]->dev_name);
#endif
            release_ezfb(valid_ezfb[i]);
#if EZFB_API_MESSAGES
            printf("DONE\n");fflush(stdout);
#endif
        }
    }
	printf("\n\n");fflush(stdout);
    exit(0);
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################

